#!/usr/bin/env python

from argparse import ArgumentParser
from math import fsum


add = lambda u, v: [i + j for i, j in zip(u, v)]        # add vectors
sub = lambda u, v: [i - j for i, j in zip(u, v)]        # subtract vectors
dot = lambda u, v: fsum(i * j for i, j in zip(u, v))    # inner product
scale = lambda a, v: [a * i for i in v]                 # vector scaling
rmul = lambda m, v: [dot(r, v) for r in m]              # matrix-vector right multiplication
tr = lambda m: list(zip(*m))                            # transposition
posdef = lambda m: [[dot(u, v) for v in t] for t in [tr(m)] for u in t]     # positively semidefinite matrix


def conjugate_gradient(a, b):
    """
    Solve system of equations Ax=b using conjugate gradients

    :param a: matrix
    :param b: vector
    :return: solution to the system
    """

    # fix system singularity
    b = rmul(tr(a), b)
    a = posdef(a)

    # conjugate gradient
    x = [0] * len(b)
    residual = sub(b, rmul(a, x))
    momentum = residual
    delta = dot(residual, residual)

    for i in range(200):
        e = dot(momentum, rmul(a, momentum))
        if e == 0:
            return x

        step = scale(delta / e, momentum)
        if dot(step, step) < 1e-28:
            return x

        x = add(x, step)
        residual = sub(b, rmul(a, x))
        delta_old = delta
        delta = dot(residual, residual)
        momentum = add(residual, scale(delta / delta_old, momentum))

    # system solution
    return x


def fit_model(data, model, length):
    """
    Fit a model to the data and the sequence values

    :param data: input data
    :param model: model complexity
    :param length: prediction length
    :return: (predicted data, total loss, model weights)
    """
    n = min(model + 1, len(data) - model)

    # build equation system
    a = [[1] + data[i:i + model] for i in range(n)]
    b = data[model:model + n]

    # solve the system
    weights = conjugate_gradient(a, b)
    prediction = data[:model]
    loss = 0

    # predict data
    while len(prediction) < length:
        x = [1] + prediction[-model:]
        h = dot(x, weights)
        prediction.append(int(round(h)))

        i = len(prediction) - 1
        loss += (data[i] - h) ** 2 if i < len(data) else 0

    # result
    return prediction, round(loss, 12), weights


def predict(data, debug=False):
    """
    Predict subsequent data using best-fitted model

    :param data: input data
    :param debug: print debug info
    :return: subsequent data based on model prediction
    """
    n = len(data)
    prediction, loss, weights = [0] * n, None, None

    # find the best prediction model
    for model in range(1, n // 2 + 1):
        try:
            prediction, loss, weights = fit_model(data, model, n * 2)
            if loss < 1e-12:
                break
        except Exception as e:
            if debug:
                print('error on', data, 'model', model)
                print(e)

    # debug
    if debug:
        print('input', data)
        print('prediction', prediction)
        print('loss', loss)
        print('weights', weights)
        print()

    return prediction[n:]


def main():
    # cli
    parser = ArgumentParser()
    parser.add_argument('input', help='input file')
    parser.add_argument('output', help='output file')
    parser.add_argument('--debug', action='store_true', help='print debug info')
    args = parser.parse_args()

    # read input
    with open(args.input, 'r') as f:
        lines = f.readlines()

    # process data
    sequences = [list(map(int, l.split(','))) for l in lines]
    predictions = [predict(s, args.debug) for s in sequences]
    lines = [','.join(map(str, p)) + '\n' for p in predictions]

    # write output
    with open(args.output, 'w') as f:
        f.writelines(lines)


if __name__ == '__main__':
    main()
